module vsmBTree
#ifndef CLANG
 use mpi
#endif
 use vsmConst
 use vsmFunction
 use vsmParam
 use vsmImage

 implicit none
 private

 type vismo_btree
   integer :: numstage
   integer :: myrank
   integer :: psize
   integer, dimension(:), allocatable :: partner
 end type vismo_btree

 type(vismo_btree), save :: vsmBTreeParam
 public vsmBinaryTreeIM, vsmFinalizeBTree, vsmInitBTree

 contains

 subroutine vsmFinalizeBTree
  deallocate(vsmBTreeParam%partner)
 end subroutine vsmFinalizeBTree

 subroutine vsmInitBTree(rank, psize)
   integer, intent(in) :: rank, psize
   integer :: i
   integer :: allocstate
   logical :: possible

   call vsmLogPrint("vsmInitBTree is called")
   call vsmCheckPowerOf2(psize, vsmBTreeParam%numstage, possible)
   vsmBTreeParam%myrank = rank
   vsmBTreeParam%psize = psize

   allocate(vsmBTreeParam%partner(vsmBTreeParam%numstage), stat=allocstate)
   if(allocstate /= 0) call vsmError("can't allocate btree partner")

   do i=1, vsmBTreeParam%numstage
     vsmBTreeParam%partner(i) = -1
   end do

   call vsmDvProcess4BT(rank, psize)

   call vsmLogPrint("vsmInitBTree ends")
 end subroutine vsmInitBTree

 subroutine vsmDvProcess4BT(rank, psize)
  integer, intent(in) :: rank, psize
  integer :: nstage, n
  integer :: NumBTProcess
  integer :: prank
  integer :: p, np, hp

  nstage = vsmBTreeParam%numstage

  NumBTProcess = 2**nstage

  if(rank >= NumBTProcess) return

  n = 0
  do
    p = 2**n
    np = NumBTProcess/p
    if(rank >= np) exit
    hp = np/2
    if(rank >= hp) then
       prank = rank - hp
    else
       prank = rank + hp
    end if

    vsmBTreeParam%partner(n+1) = prank
    if(vsmDbg) then
       call vsmLogPrint("binary swap 4 solid : stage = ", nstage)
       call vsmLogPrint("binary swap 4 solid : myrank = ", rank)
       call vsmLogPrint("binary swap 4 solid : parter rank = ", prank)
    end if
    n = n + 1
    if(nstage == n) exit
  end do
 end subroutine vsmDvProcess4BT

  subroutine vsmBinaryTreeIM(cam, vis, tim)
   integer, intent(in) :: cam, vis, tim
   integer :: comm, myrank, psize, ierr
#ifndef CLANG
   integer, dimension(mpi_status_size) :: stat
#endif
   integer :: pixels, isize
   integer :: i, st_pixel, ed_pixel, prank
   integer :: nbtproc, exproc, hp
   integer :: tag, src
   integer, dimension(2) :: spixl

   character(4) :: viskind, camnum, rank 
   character(7) :: timestep

   if(vsmPCmode /= 0) call vsmError("BinaryTreeIm is called though pcmode is not 0")
   tag = 0
   src = 0
   call vsmLogPrint("----- Binary Tree")
   call vsmGetMPIVars(comm, myrank, psize)

   nbtproc = 2**vsmBTreeParam%numstage
   isize = vsmImageData%rimgsize*5
#ifndef CLANG
   call mpi_barrier(comm, ierr)
#else
   call vsmmpibarrier()
#endif

   if(psize > nbtproc) then

     exproc = psize - nbtproc

     if(myrank < exproc) then
       prank = myrank + nbtproc
#ifndef CLANG
       call mpi_recv(vsmImageData%image_r, isize, mpi_real, prank, tag, comm, stat, ierr)
#else
       call vsmmpirecvfloat(vsmImageData%image_r, isize, prank)
#endif
       call vsmSolidBlending(1, vsmImageData%rimgsize)
     else if(myrank > nbtproc-1) then
       prank = myrank - nbtproc
#ifndef CLANG
       call mpi_send(vsmImageData%image, isize, mpi_real, prank, tag, comm, ierr)
#else
       call vsmmpisendfloat(vsmImageData%image, isize, prank)
#endif
     end if

   end if

   if(myrank < nbtproc) then

     st_pixel = 1
     ed_pixel = vsmImageData%rimgsize !vsmImageData%width * vsmImageData%height ! send receive data*5

     call vsmCheckImageL(st_pixel, ed_pixel)
     do i=1, vsmBTreeParam%numstage
       call vsmLogPrint("binary tree for solid stage = ", i)
       prank = vsmBTreeParam%partner(i)

       if(prank == -1) cycle

       call vsmClearRImage
       hp = 2**(vsmBTreeParam%numstage -i )!2**(vsmBTreeParam%numstage -i+1)/2
       if(myrank < hp) then
         call vsmLogPrint("binary tree  recv from= ", prank)
#ifndef CLANG
         call mpi_recv(vsmImageData%image_r, isize, mpi_real, prank, tag, comm, stat, ierr)
#else
         call vsmmpirecvfloat(vsmImageData%image_r, isize, prank)
#endif
         call vsmSolidBlending(st_pixel, ed_pixel)
         call vsmCheckImageL(st_pixel, ed_pixel)
       else
         call vsmLogPrint("binary tree  send to= ", prank)
#ifndef CLANG
         call mpi_send(vsmImageData%image, isize, mpi_real, prank, tag, comm, ierr)
#else
         call vsmmpisendfloat(vsmImageData%image, isize, prank)
#endif
       end if

     end do

   end if
   isize = vsmImageData%rimgsize*5
#ifndef CLANG
   call mpi_bcast(vsmImageData%image, isize, mpi_real, src, comm, ierr)
#else
   call vsmmpibcastfloat(vsmImageData%image, isize, src)
#endif
   !----
   call vsmLogPrint("----- Binary Tree for Solid ends")
 end subroutine vsmBinaryTreeIM

end module vsmBTree

